// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the GNU General Public License version 2 ("the License").
// You may not use this file except in compliance with the License.

package org.xwt.shoehorn3;

import java.applet.*;
import java.util.*;
import java.lang.reflect.*;
import java.io.*;
import java.net.*;
import java.util.zip.*;
import java.awt.*;
import java.math.*;

/** This class is XWT's presence on the user's computer; it must be run as trusted code */
public class ShoeHorn extends Applet {

    // Startup Phase ////////////////////////////////////////////////////////////////////
    private String build = null;
    public ShoeHorn() { log("*** constructor invoked for " + this.getClass().getName()); }

    public final String getParameter(String arg) { return super.getParameter(arg); }
    public final void main(String[] s) { new ShoeHorn().start(); }
    private void log(String s) { System.out.println(s); }

    /** this just ensures that we are running with full privileges */
    public final void start() {
        build = getParameter("build");
        new Thread() { public void run() { 
            log("ShoeHorn thread spawned");
            try {

                if (System.getProperty("java.vendor", "").startsWith("Netscape")) {
                    log("Detected Navigator 4.x");
                    Method m = Class.forName("netscape.security.PrivilegeManager").getMethod("enablePrivilege", new Class[] { String.class });
                    m.invoke(null, new Object[] { "MarimbaInternalTarget" });
                    m.invoke(null, new Object[] { "UniversalExecAccess" });
                    m.invoke(null, new Object[] { "UniversalPropertyRead" });
                    go();

                } else if (System.getProperty("java.vendor", "").startsWith("Microsoft")) {
                    //com.ms.security.PolicyEngine.assertPermission(com.ms.security.PermissionID.SYSTEM);
                    Class permissionIdClass = Class.forName("com.ms.security.PermissionID");
                    Object o = permissionIdClass.getField("SYSTEM").get(null);
                    Method m = Class.forName("com.ms.security.PolicyEngine").getMethod("assertPermission", new Class[] { permissionIdClass });
                    m.invoke(null, new Object[] { o });
                    go();

                } else {
                    log("Detected non-Navigator JVM");
                    Method m = Class.forName("org.xwt.shoehorn3.ShoeHorn$Java12").getMethod("run", new Class[] { Object.class });
                    m.invoke(null, new Object[] { ShoeHorn.this });
                }
            } catch (Throwable e) {
                if (e instanceof InvocationTargetException) e = ((InvocationTargetException)e).getTargetException();
                e.printStackTrace();
                update(-1.0, "Error; please check the Java console");
            }
        } }.start();
    }

    /** ask Java Plugin for privs */
    private static class Java12 {
        public static void run(final Object a) {
            java.security.AccessController.doPrivileged(new java.security.PrivilegedAction() {
                    public Object run() {
                        ((ShoeHorn)a).go();
                        return null;
                    }
                });
        }
    }


    // Implantation ////////////////////////////////////////////////////////////////////

    /** inserts the required entries into the user's ~/.java.policy */
    private void modifyPolicyFile() throws IOException {
        log("Adjusting ~/.java.policy");
        File policy = new File(System.getProperty("user.home") + File.separatorChar + ".java.policy");
        if (policy.exists()) {
            BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(policy)));
            String s = null;
            while((s = br.readLine()) != null)
                if (s.startsWith("// XWT_MARKER:")) {
                    log("Java policy file has already been adjusted");
                    return;
                }
        }
        FileOutputStream fos = new FileOutputStream(policy.getAbsolutePath(), true);
        PrintWriter pw = new PrintWriter(new OutputStreamWriter(fos));
        pw.println("");
        pw.println("// XWT_MARKER: this line and the following two grant blocks are required for XWT; DO NOT REMOVE THEM.");
        pw.println("grant {");
        pw.println("    permission java.io.FilePermission \"${user.home}${/}.xwt${/}shoehorn.jar\", \"read\";");
        pw.println("};");
        pw.println("grant codebase \"file:${user.home}${/}.xwt${/}shoehorn.jar\" {");
        pw.println("    permission java.security.AllPermission;");
        pw.println("};");
        pw.println("// END_XWT_MARKER");
        pw.println("");
        pw.flush();
        pw.close();
    }

    /** read ourselves out of the resources and write a jarfile to some trusted place */
    private void implantSelf() throws IOException {
        InputStream manifest = getClass().getClassLoader().getResourceAsStream("META-INF/manifest.mf");
        log("my classloader is " + getClass().getClassLoader().getClass().getName());
        ClassLoader loader = getClass().getClassLoader();
        if (manifest == null || loader == null ||
            (loader.getClass().getName().indexOf("Applet") == -1 && loader.getClass().getName().indexOf("Plugin") == -1))
            return;
        BufferedReader br = new BufferedReader(new InputStreamReader(manifest));
        Vector entries = new Vector();
        String s = null;
        while((s = br.readLine()) != null)
            if (s.startsWith("Name: "))
                entries.addElement(s.substring(6));

        String ext_dirs = System.getProperty("java.ext.dirs");
        log("java.ext.dirs = " + ext_dirs);
        ext_dirs = ext_dirs + File.pathSeparatorChar + System.getProperty("user.home") + File.separatorChar + ".xwt";
        StringTokenizer st = new StringTokenizer(ext_dirs, File.pathSeparatorChar + "");
        while(st.hasMoreTokens()) {
            String dir = st.nextToken();
            new File(dir).mkdirs();
            try {
                // we have to modify the policy file BEFORE implanting in ~/.xwt to ensure that the policy mods work
                if (!st.hasMoreTokens()) modifyPolicyFile();
                implantInDirectory(dir, entries);
                return;
            } catch (IOException e) {
                log("Failed to implant in " + dir + " due to " + e);
            }
        }
        log("Failed to implant self!");
    }

    private void implantInDirectory(String dir, Vector entries) throws IOException {
        File f = new File(dir + File.separatorChar + "shoehorn.tmp");
        ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(f));
        for(int i=0; i<entries.size(); i++) {
            zos.putNextEntry(new ZipEntry(entries.elementAt(i).toString()));
            InputStream is = this.getClass().getClassLoader().getResourceAsStream(entries.elementAt(i).toString());
            byte[] buf = new byte[1024];
            while(true) {
                int read = is.read(buf, 0, buf.length);
                if (read == -1) break;
                zos.write(buf, 0, read);
            }
            is.close();
        }
        zos.close();
        f.renameTo(new File(dir + File.separatorChar + "shoehorn.jar"));
        log("Succeeded in implanting in " + dir);
    }


    // Startup Phase ////////////////////////////////////////////////////////////////////

    public final void go() {
        try {
            update(0.0, "");
            implantSelf();
            
            File file;
            String os_name = System.getProperty("os.name", "");
            log("os.name == " + os_name);
            Vector command = new Vector();
            
            String arch = null;
            if (os_name.indexOf("Linux") != -1) {
                arch = new BufferedReader(new InputStreamReader(Runtime.getRuntime().exec("/bin/uname -m").getInputStream())).readLine();
                log("arch is " + arch);
            }
            
            if (os_name.indexOf("Linux") != -1 && arch.charAt(0) == 'i' && arch.substring(2, 4).equals("86")) {
                file = fetch("xwt-" + build + ".linux.gz");
                safeWaitFor(Runtime.getRuntime().exec("/bin/chmod +x " + file.getAbsolutePath()));
                command.addElement(file.getAbsolutePath());
                
            } else if (os_name.indexOf("Windows") != -1) {
                file = fetch("xwt-" + build + ".cab");
                command.addElement(file.getAbsolutePath());
                
            } else {
                file = fetch("xwt-" + build + ".jar");
                command.addElement(findJvmBinary());
                command.addElement("-jar");
                command.addElement(file.getAbsolutePath());
            }
            
            if ("true".equals(getParameter("showrenders"))) command.addElement("-s");
            if ("true".equals(getParameter("verbose"))) command.addElement("-v");
            if (getParameter("log") != null) {
                command.addElement("-l");
                command.addElement(getParameter("log"));
            }
            command.addElement(getParameter("xwar"));
            spawn(command);
        } catch (Exception e) {
            update(-1.0, "Error; please check the Java console");
            e.printStackTrace();
        }
    }

    /** searches for the JVM binary in the usual places */
    private String findJvmBinary() throws IOException {
        String jvmBinary = null;

        // HACK: prefer jdk1.3 on OSX; it has better fonts
        if (new File("/System/Library/Frameworks/JavaVM.framework/Versions/1.3.1/Commands/java").exists())
            jvmBinary = "/System/Library/Frameworks/JavaVM.framework/Versions/1.3.1/Commands/java";

        // check JAVA_HOME
        String javaHome = getEnv("JAVA_HOME");
        if (jvmBinary == null && javaHome != null && !javaHome.equals("")) {
            jvmBinary = javaHome + File.separatorChar + "bin" + File.separatorChar + "java";
            if (!new File(jvmBinary).exists()) jvmBinary = null;
        }
        
        // check common locations
        if (jvmBinary == null)
            for(int i=0; i<commonJavaLocations.length; i++)
                if (new File(commonJavaLocations[i]).exists()) {
                    jvmBinary = commonJavaLocations[i];
                    break;
                }
        
        // check PATH
        if (jvmBinary == null) {
            String path = getEnv("PATH");
            StringTokenizer st = new StringTokenizer(path, File.pathSeparatorChar + "");
            while(st.hasMoreTokens()) {
                String s = st.nextToken();
                if (new File(s + File.separatorChar + "java").exists()) {
                    jvmBinary = s + File.separatorChar + "java";
                    break;
                }
            }
        }
        
        // check ${java.home}
        javaHome = System.getProperty("java.home");
        if (jvmBinary == null && javaHome != null && !javaHome.equals("")) {
            jvmBinary = javaHome + File.separatorChar + "bin" + File.separatorChar + "java";
            if (!new File(jvmBinary).exists()) jvmBinary = null;
        }
        
        if (jvmBinary == null)
            throw new Error("couldn't find JVM binary! JAVA_HOME=" + getEnv("JAVA_HOME") + " PATH=" + getEnv("PATH"));
        else
            return jvmBinary;
    }

    private void spawn(Vector command) throws IOException {
        String proxy = detectProxy();
        log("proxy settings: " + proxy);

        String[] command_vec;
        command.copyInto(command_vec = new String[command.size()]);
        log("executing:");
        for(int i=0; i<command_vec.length; i++) log("    \"" + command_vec[i] + "\"");

        Process p;
        if (proxy == null) {
            p = Runtime.getRuntime().exec(command_vec);
        } else if (System.getProperty("os.name").toLowerCase().indexOf("windows") != -1) {
            // hrm, win32 doesn't like it when we probe the environment; that doesn't matter, since the
            // win32 environ vars are pretty much useless anyways.
            p = Runtime.getRuntime().exec(command_vec, new String[] { proxy });
        } else {
            p = Runtime.getRuntime().exec(command_vec, dumpEnv(proxy));
        }
        BufferedReader stderr = new BufferedReader(new InputStreamReader(p.getErrorStream()));
        
        if (System.getProperty("os.name").toLowerCase().indexOf("windows") != -1) {
            update(1.0, "XWT Loaded");
        } else {
            String s = stderr.readLine();
            update(1.0, "XWT Loaded");
            while(s != null) {
                log(s);
                s = stderr.readLine();
            }
        }
        log("exiting...");
    }


    // Utility Functions /////////////////////////////////////////////////////////////////////
    
    /** Voodoo to extract the proxy settings from Sun's Java Plugin */
    private String detectProxy() {
        try {
            Vector ret = new Vector();
            
            Class PluginProxyHandler = Class.forName("sun.plugin.protocol.PluginProxyHandler");
            Method getDefaultProxyHandler = PluginProxyHandler.getMethod("getDefaultProxyHandler", new Class[] { });
            Object proxyHandler = getDefaultProxyHandler.invoke(null, new Object[] { });
            
            Class ProxyHandler = Class.forName("sun.plugin.protocol.ProxyHandler");
            Method getProxyInfo = ProxyHandler.getMethod("getProxyInfo", new Class[] { URL.class });
            Object proxyInfo = getProxyInfo.invoke(proxyHandler, new Object[] { new URL("http://www.xwt.org") });
            
            Class ProxyInfo = Class.forName("sun.plugin.protocol.ProxyInfo");
            
            if (((Boolean)ProxyInfo.getMethod("isProxyUsed", new Class[] { }).invoke(proxyInfo, new Object[] { })).booleanValue())
                return "http_proxy=" +
                    (String)ProxyInfo.getMethod("getProxy", new Class[] { }).invoke(proxyInfo, new Object[] { }) + ":" +
                    ((Integer)ProxyInfo.getMethod("getPort", new Class[] { }).invoke(proxyInfo, new Object[] { })).intValue();

            if (((Boolean)ProxyInfo.getMethod("isSocksUsed", new Class[] { }).invoke(proxyInfo, new Object[] { })).booleanValue())
                return "socks_proxy=" +
                    (String)ProxyInfo.getMethod("getSocksProxy", new Class[] { }).invoke(proxyInfo, new Object[] { }) + ":" +
                    ((Integer)ProxyInfo.getMethod("getSocksPort", new Class[] { }).invoke(proxyInfo, new Object[] { })).intValue();

            return null;
            
        } catch (Throwable e) {
            log("exception while querying sun.plugin.protocol.PluginProxyHandler: " + e);
            return null;
        }
    }

    // The Netscape JVM has a bug which causes waitFor() to lock up, so we wait no more than 2000ms
    private void safeWaitFor(final Process p) {
        final Object o = new Object();
        new Thread() {
                public void run() {
                    try {
                        p.waitFor();
                    } catch (InterruptedException e) { }
                    synchronized(o) { o.notify(); }
                }
            }.start();
        try {
            synchronized(o) { o.wait(2000); }
        } catch (InterruptedException e) { }
    }

    /** fetches a file from the distribution site, writing it to the appropriate place */
    private File fetch(String filename) throws IOException {
        String tmpdir = System.getProperty("user.home") + File.separatorChar + ".xwt";
        new File(tmpdir).mkdirs();
        URL u = new URL("http://dist.xwt.org/" + filename);
        if (filename.endsWith(".gz")) filename = filename.substring(0, filename.length() - 3);
        if (filename.endsWith(".cab")) filename = filename.substring(0, filename.length() - 4) + ".exe";
        File target = new File(tmpdir + File.separatorChar + filename);
        if (target.exists()) return target;
        loadFromURL(u, target, filename);
        return target;
    }

    /** loads a file from a url, verifying that it was properly signed */
    private void loadFromURL(URL u, File target, String filename) throws IOException {

        final URLConnection uc = u.openConnection();
        final int signatureLength = 128;
        final byte[] signature = new byte[signatureLength];

        InputStream is = uc.getInputStream();
        final int contentLength = uc.getContentLength();

        // display the progress indicator as we go
        InputStream dis = new FilterInputStream(is) {
                int total = 0;
                public int read() throws IOException {
                    int ret = super.read();
                    if (ret != -1) total++;
                    double loaded = ((double)total) / ((double)uc.getContentLength());
                    update(loaded, "Loading XWT: " + ((int)Math.ceil(loaded * 100)) + "%");
                    return ret;
                }
                public int read(byte[] buf, int off, int len) throws IOException {
                    int ret = super.read(buf, off, len);
                    if (ret != -1) total += ret;
                    double loaded = ((double)total) / ((double)uc.getContentLength());
                    update(loaded, "Loading XWT: " + ((int)Math.ceil(loaded * 100)) + "%");
                    return ret;
                }
            };

        // decompress if needed
        if (u.toString().endsWith(".cab")) dis = CAB.getFileInputStream(dis, "xwt-" + build + ".exe");
        else if (u.toString().endsWith(".gz")) dis = new GZIPInputStream(dis);

        // digest and copy the file to target.tmp, omitting the last signatureLength bytes
        SHA1Digest sha1 = new SHA1Digest();
        OutputStream fos = new DigestOutputStream(new FileOutputStream(target + ".tmp"), sha1);
        byte[] buf = new byte[1024 * 128];
        int numread = 0;
        while(true) {
            while(numread < buf.length) {
                int read = dis.read(buf, numread, buf.length - numread);
                if (read == -1) break;
                numread += read;
            }
            if (numread < buf.length) {
                System.arraycopy(buf, numread - signatureLength, signature, 0, signatureLength);
                fos.write(buf, 0, numread - signatureLength);
                break;
            } else {
                fos.write(buf, 0, buf.length - signatureLength);
                System.arraycopy(buf, buf.length - signatureLength, buf, 0, signatureLength);
                numread = signatureLength;
            }
        }
        fos.close();

        // construct our hash
        byte[] hash = new byte[sha1.getDigestSize()];
        sha1.doFinal(hash, 0);

        // decrypt the signature hash
        DERInputStream deris = new DERInputStream(new ByteArrayInputStream(Base64.decode(XWT_foundation_public_key)));
        SubjectPublicKeyInfo pki = new SubjectPublicKeyInfo((DERConstructedSequence)deris.readObject());
        RSAPublicKeyStructure rsa_pks = new RSAPublicKeyStructure((DERConstructedSequence)pki.getPublicKey());
        BigInteger modulus = rsa_pks.getModulus();
        BigInteger exponent = rsa_pks.getPublicExponent();
        AsymmetricBlockCipher rsa = new PKCS1Encoding(new RSAEngine());
        rsa.init(false, new RSAKeyParameters(false, modulus, exponent));

        // compare
        try {
            byte[] decryptedHash = rsa.processBlock(signature, 0, signature.length);
            for(int i=0; i<hash.length; i++)
                if (decryptedHash[i] != hash[i]) {
                    new File(target + ".tmp").delete();
                    throw new Error("invalid signature on " + filename);
                }
        } catch (InvalidCipherTextException e) {
            new File(target + ".tmp").delete();
            throw new Error(e);
        }

        // good to go; rename the file
        new File(target + ".tmp").renameTo(target);
    }

    /** retrieves an environment variable */
    private static String getEnv(String key) throws IOException {
        Process p = Runtime.getRuntime().exec(new String[] { "/bin/sh", "-c", "echo $" + key });
        BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()));
        return br.readLine();
    }

    /** dumps the environment plus an additional variable */
    private static String[] dumpEnv(String newvar) throws IOException {
        Vector v = new Vector();
        String os = System.getProperty("os.name").toLowerCase();
        Process p;
        if (os.indexOf("windows 9") > -1) {
            p = Runtime.getRuntime().exec("command.com /c set");
        } else if ((os.indexOf("nt") > -1) || (os.indexOf("windows 2000") > -1) ) {
            p = Runtime.getRuntime().exec("cmd.exe /c set");
        } else {  
            p = Runtime.getRuntime().exec("env");
        }
        BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()));
        String s;
        while ((s = br.readLine()) != null) v.addElement(s);
        v.addElement(newvar);
        String[] ret;
        v.copyInto(ret = new String[v.size()]);
        return ret;
    }



    // Applet Painting Functions ////////////////////////////////////////////////////////

    private Image backbuffer = null;
    public final void paint(Graphics g) { if (backbuffer != null) g.drawImage(backbuffer, 0, 0, null); }

    public final Graphics getGraphics() { return super.getGraphics(); }
    public final Dimension getSize() { return super.getSize(); }

    private void update(double loaded, String text) {
        // indicates we should paint ourselves
        Graphics g2 = getGraphics();
        String s = text;
        if (backbuffer == null || backbuffer.getWidth(null) != getSize().width || backbuffer.getHeight(null) != getSize().height)
            backbuffer = createImage(getSize().width, getSize().height);
        if (backbuffer == null) return;
        Graphics g = backbuffer.getGraphics();

        g.setColor(loaded < 0 ? Color.red : Color.blue);
        loaded = Math.abs(loaded);

        int w = (int)((double)getSize().width * loaded);
        g.fillRect(0, 0, w, getSize().height);
        g.setColor(Color.darkGray);
        g.fillRect(w, 0, getSize().width - w, getSize().height);

        Font f = new Font("Sans-serif", Font.BOLD, 12);
        FontMetrics fm = Toolkit.getDefaultToolkit().getFontMetrics(f);
        g.setFont(f);

        int x = (getSize().width - fm.stringWidth(s)) / 2;
        int y = ((getSize().height - fm.getMaxAscent() - fm.getMaxDescent()) / 2) + fm.getMaxAscent();
        g.setColor(Color.white);
        g.drawString(s, x, y);

        if (g2 != null) {
            g2.setClip(0, 0, getSize().width, getSize().height);
            g2.drawImage(backbuffer, 0, 0, null);
        }
    }


    // Misc Constants ///////////////////////////////////////////////////////////

    private static final String XWT_foundation_public_key =
        "MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCoVweDZ+h3jcN2Qz1YwWJR8gVF0m/IE" +
        "BasLIcvey9y1VkyN9jY6b9qm/Z2UoVtvAlgezd4CsJedUCIMe7uURyHGnqI4MLrozxLz3" +
        "zqx5EYChsJt+Ju/f44KYnMx7upUQ4irfxOj6RpHy3E5GbW4XO96WwFlOuaR8+HRwKCXGP" +
        "QvQIDAQAB";

    private static final String[] commonJavaLocations = new String[] {
        "/usr/bin/java",
        "/usr/java/bin/java",
        "/usr/local/bin/java",
        "/usr/local/java/bin/java",
        "/usr/lib/j2sdk1.4/bin/java",
        "/usr/lib/j2sdk1.3/bin/java",
        "/usr/lib/j2sdk1.2/bin/java"
    };

}




/*

/// NSJVM Notes ///////////////////////////////////////////////////////////////////

  PrivilegeManager.enablePrivilege("MarimbaInternalTarget");
  PrivilegeManager.enablePrivilege("UniversalSystemClipboardAccess");
  netscape.security.PrivilegeManager.enablePrivilege("UniversalConnect"); (sockets)
  netscape.security.PrivilegeManager.enablePrivilege("UniversalSystemClipboardAccess");
  netscape.security.PrivilegeManager.enablePrivilege("UniversalTopLevelWindow");

  - Netscape's ClassLoader.getResource() is broken, see http://developer.netscape.com/docs/technote/java/getresource/getresource.html
  - this will fix it: netscape.security.PrivilegeManager.enablePrivilege("UniversalPropertyRead");

  - If you create a classloader, include
        public boolean checkMatchPrincipalAlways(int i) {
            return ((AppletClassLoader)this.getClass().getClassLoader()).checkMatchPrincipalAlways(i);
        }
        
  - protected int _modifiersToButtonNumber(int modifiers) {
        if ((modifiers & (InputEvent.BUTTON1_MASK & 15)) == (InputEvent.BUTTON1_MASK & 15)) return 1;
        if ((modifiers & (InputEvent.BUTTON2_MASK & 15)) == (InputEvent.BUTTON2_MASK & 15)) return 3;
        if ((modifiers & (InputEvent.BUTTON3_MASK & 15)) == (InputEvent.BUTTON3_MASK & 15)) return 2;


/// MSJVM Notes ///////////////////////////////////////////////////////////////////

- To sniff the JVM, check if (window.clientInformation.javaEnabled()), or if
  (navigator.javaEnabled()).  See also http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dndetect/html/detectvm.asp

- MSJVM note: you have to write to the getGraphics() of a Frame before you can setIconImage() it.

- How to create trusted classes within an already-running MSJVM:

    class MicrosoftClassLoader extends SecurityClassLoader {
        static PermissionSet ps;  // a PermissionSet that allows <i>everything</i>
        static {
                PermissionDataSet pds = new PermissionDataSet();
                    IPermission perm = new IPermission() {
                    public void check(Object request) { }
                    public IPermission combine(IPermission other) { return this; }
                    public IPermission copy() { return this; }
                    public int compareSet(Object o) { return SetComparison.DISJOINT; }
                };
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.ClientStoragePermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.ExecutionPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.FileIOPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.MultimediaPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.NetIOPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.PrintingPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.PropertyPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.ReflectionPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.SecurityPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.SystemStreamsPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.ThreadPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.UIPermission"), perm);
            pds.add(PolicyEngine.permissionNameToID("com.ms.security.permissions.UserFileIOPermission"), perm);
            ps = new PermissionSet(pds);
        }
        public Class defineClass(String name, byte[] b) {
            Class c = super.defineClass(name, b, 0, b.length, ps, com.ms.security.PolicyEngine.getPrincipalOfClass(Resources.class));
            if (name.startsWith("net")) System.out.println("defineClass " + name + " yielded " + c);
            // for some reason, the MSJVM doesn't request resolves properly, so we have to do it manually on every class-load
            resolveClass(c);
            return c;
        }

- How to inhibit background-clearing on the MSJVM

        // Used to ensure that getPeer() doesn't go off in an infinite loop
        boolean ok = false;

        // The MSJVM needs us to occasionally falsify getPeer() in order to thwart unneeded background-clearing
        public ComponentPeer getPeer() {
            if (Thread.currentThread() == Platform.fastPathThread) return super.getPeer();
            if (ok) return super.getPeer();

            // to prevent recursive stack-dumping... =)
            ok = true;
            Dimension d = getSize();
            ok = false;
            if (last != null && last.width == d.width && last.height == d.height)
                return super.getPeer();

            String s = com.ms.wfc.util.Debug.getStackTraceText();
            for(int i = s.indexOf('d'); i != -1; i = s.indexOf('d', i + 1)) {
                if (s.regionMatches(i, "doClearAndPaint", 0, 15)) {
                    last = getSize(); 
                    update(null);
                    return null;
                }
            }
            return super.getPeer();
        }

*/
